﻿using System;
using System.Collections.Generic;
using System.Linq;
using System.Net;
using System.Net.Sockets;
using System.Windows.Forms;
using PushSharp;
using PushSharp.Android;
using PushSharp.Common;

namespace EGRemoteStudio
{
    public class DeviceHandler : IDisposable
    {
        private readonly DeviceList deviceList;
        private RegAsyncTcpServer tcpServer;
        private readonly DeviceListManager manager;
        private readonly List<NetworkEventreceiver> receiverList;

        private const int FIRST_PORT = 6789;
        private const int LAST_PORT = 6900;

        private const int SUPPORTED_DEVICE_VERSION = 19;

        private readonly GcmPushService push;
        private const string PROJECT_ID = "180673848444";
        private const string SENDER_AUTH = "AIzaSy123WhTfRg8394VSh123dX123RWXDZxg6c";
        private const string APP_PACKAGE_NAME = "eu.rml.lab.egremotestudio";

        public event NetworkEventreceiver.NetworkEventHandler OnNetworkEvent;
        public event ListenHandler OnListenEvent;
        public delegate void ListenHandler(object sender, ListenEventArgs e);

        public class ListenEventArgs
        {
            public int Port { get; set; }
        }

        public DeviceList DeviceList
        {
            get { return deviceList; }
        }

        public DeviceHandler()
        {
            manager = new DeviceListManager();
            deviceList = new DeviceList();
            deviceList.DeviceListChanged += deviceList_DeviceListChanged;
            receiverList = new List<NetworkEventreceiver>();

            /* initialize push service */
            push = new GcmPushService(new GcmPushChannelSettings(PROJECT_ID, SENDER_AUTH, APP_PACKAGE_NAME));

            //Wire up the events
            push.Events.OnDeviceSubscriptionExpired += Events_OnDeviceSubscriptionExpired;
            push.Events.OnDeviceSubscriptionIdChanged += Events_OnDeviceSubscriptionIdChanged;
            push.Events.OnChannelException += Events_OnChannelException;
            push.Events.OnNotificationSendFailure += Events_OnNotificationSendFailure;
            push.Events.OnNotificationSent += Events_OnNotificationSent;
            push.Events.OnChannelCreated += Events_OnChannelCreated;
            push.Events.OnChannelDestroyed += Events_OnChannelDestroyed;
        }

        public void StartListener(int port)
        {
            /* listen for new device registrations */
            tcpServer = new RegAsyncTcpServer(IPAddress.Any, port);
            tcpServer.OnDeviceRegister += tcpServer_OnDeviceRegister;
            tcpServer.Start();

            /* no exception? okay, let's propagate the new used port */
            if (OnListenEvent != null)
                OnListenEvent(this, new ListenEventArgs() { Port = port });
        }

        public void StopListener()
        {
            /* no new clients may register */
            tcpServer.OnDeviceRegister -= tcpServer_OnDeviceRegister;
            tcpServer.Stop();
        }

        void deviceList_DeviceListChanged(object sender, DeviceList.DeviceListChangedArgs e)
        {
            /* now store the device list */
            manager.WriteDeviceListFile(DeviceList);
        }

        /// <summary> Restores the devices from the isolated storage file </summary>
        public void RestoreDevices()
        {
            foreach (Device device in manager.ReadDeviceListFile())
            {
                /* we try to use the port defined in the file, but if this fails we try the next port */
                int port = device.Port;

                while (port < LAST_PORT)
                    try
                    {
                        NetworkEventreceiver egNetworkEventreceiver = new NetworkEventreceiver(IPAddress.Any, device.Port);
                        egNetworkEventreceiver.Start();
                        egNetworkEventreceiver.OnNetworkEvent += egNetworkEventreceiver_OnNetworkEvent;
                        egNetworkEventreceiver.Device = device;
                    
                        receiverList.Add(egNetworkEventreceiver);
                        deviceList.Add(device);
                        break;
                    }
                    catch (SocketException)
                    {
                        port++;
                    }
            }
        }

        /// <summary> Sends a message to the specified reg id. </summary>
        /// <param name="regId">The reg id.</param>
        /// <param name="ntwEvent">The NTW event.</param>
        /// <param name="ntwpayload">The ntwpayload.</param>
        public void SendMessage(string regId, string ntwEvent, string ntwpayload)
        {
            // TODO: perhaps the timetolive should be settable by user ...
            // atm the timetolive is set to 0, so google will try to send it immediately (no delay), but if this is
            // not possible the message will be discarded!!!
            GcmNotification nf = NotificationFactory.AndroidGcm()
                .ForDeviceRegistrationId(
                    regId)
                .WithCollapseKey("NONE")
                .WithJson("{\"event\":\"" + ntwEvent + "\",\"payload\":\"" + ntwpayload + "\"}");
            nf.TimeToLive = 0;
            push.QueueNotification(nf);
        }

        /// <summary> Handles the OnNetworkEvent event of the egNetworkEventreceiver control. </summary>
        /// <param name="sender">The source of the event.</param>
        /// <param name="e">The <see cref="NetworkEventreceiver.NetworkEventArgs" /> instance containing the event data.</param>
        void egNetworkEventreceiver_OnNetworkEvent(object sender, NetworkEventreceiver.NetworkEventArgs e)
        {
            NetworkEventreceiver receiver = sender as NetworkEventreceiver;

            if (receiver != null)
                SendMessage(receiver.Device.RegId, e.ntwEvent, e.ntwPayload);

            if (OnNetworkEvent != null)
                OnNetworkEvent(sender, e);
        }

        /// <summary> Called when a new device has registered. </summary>
        /// <param name="sender">The sender.</param>
        /// <param name="e">The e.</param>
        void tcpServer_OnDeviceRegister(object sender, RegAsyncTcpServer.DeviceRegisterArgs e)
        {
            /* check if there is already a device registered with the given id, if so just return */
            if (deviceList.Any(d => d.RegId == e.regId))
                return;

            /* search for the next unused port for the EG communication */
            int port = FIRST_PORT;

            while (port < LAST_PORT)
            try
            {
                NetworkEventreceiver egNetworkEventreceiver = new NetworkEventreceiver(IPAddress.Any, port);
                egNetworkEventreceiver.Start();
                egNetworkEventreceiver.OnNetworkEvent += egNetworkEventreceiver_OnNetworkEvent;

                Device device = new Device(e.regId, e.model, e.manufacturer, e.appVersion, port);
                DeviceList.Add(device);

                if (SUPPORTED_DEVICE_VERSION > e.appVersion)
                    MessageBox.Show(
                        "The registered device may not be fully supported, please update your version of EGRemote!",
                        "Please update EGRemote", MessageBoxButtons.OK, MessageBoxIcon.Warning);
                else if (SUPPORTED_DEVICE_VERSION < e.appVersion)
                    MessageBox.Show(
                        "The registered device may not be fully supported, please update your version of EGRemote Studio (http://goo.gl/Kdc89)!",
                        "Please update EGRemote Studio", MessageBoxButtons.OK, MessageBoxIcon.Warning);

                egNetworkEventreceiver.Device = device;
                receiverList.Add(egNetworkEventreceiver);
                break;
            }
            catch (SocketException)
            {
                port++;
            }
        }

        /// <summary> The device subscription id changed. </summary>
        /// <param name="platform">The platform.</param>
        /// <param name="oldDeviceInfo">The old device info.</param>
        /// <param name="newDeviceInfo">The new device info.</param>
        /// <param name="notification">The notification.</param>
        private void Events_OnDeviceSubscriptionIdChanged(PlatformType platform, string oldDeviceInfo, string newDeviceInfo, Notification notification)
        {
            foreach (Device device in deviceList)
                if (device.RegId == oldDeviceInfo)
                {
                    device.ChangeRegId(newDeviceInfo);
                    /* store devicelist containing the new regId */
                    manager.WriteDeviceListFile(deviceList);
                }

            //Currently this event will only ever happen for Android GCM
            Console.WriteLine("Device Registration Changed:  Old-> " + oldDeviceInfo + "  New-> " + newDeviceInfo);
        }

        private void Events_OnNotificationSent(Notification notification)
        {
            Console.WriteLine("Sent: " + notification.Platform.ToString() + " -> " + notification);
        }

        private void Events_OnNotificationSendFailure(Notification notification, Exception notificationFailureException)
        {
            Console.WriteLine("Failure: " + notification.Platform.ToString() + " -> " + notificationFailureException.Message + " -> " + notification);
        }

        private void Events_OnChannelException(Exception exception, PlatformType platformType, Notification notification)
        {
            Console.WriteLine("Channel Exception: " + platformType.ToString() + " -> " + exception);
        }

        private void Events_OnDeviceSubscriptionExpired(PlatformType platform, string deviceInfo, Notification notification)
        {
            foreach (Device device in deviceList)
                if (device.RegId == deviceInfo)
                    deviceList.Remove(device);

            Console.WriteLine("Device Subscription Expired: " + platform.ToString() + " -> " + deviceInfo);
        }

        private void Events_OnChannelDestroyed(PlatformType platformType, int newChannelCount)
        {
            Console.WriteLine("Channel Destroyed for: " + platformType.ToString() + " Channel Count: " + newChannelCount);
        }

        private void Events_OnChannelCreated(PlatformType platformType, int newChannelCount)
        {
            Console.WriteLine("Channel Created for: " + platformType.ToString() + " Channel Count: " + newChannelCount);
        }

        /// <summary> Performs application-defined tasks associated with freeing, releasing, or resetting unmanaged resources. </summary>
        public void Dispose()
        {
            foreach (NetworkEventreceiver receiver in receiverList)
            {
                receiver.OnNetworkEvent -= egNetworkEventreceiver_OnNetworkEvent;
                receiver.Stop();
            }
            deviceList.DeviceListChanged -= deviceList_DeviceListChanged;
            
            /* shutdown push service */
            push.Stop(true);
        }
    }
}
