﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.Net;
using System.Net.Sockets;
using System.Security.Cryptography;
using System.Text;

namespace EGRemoteStudio
{
    /// <summary>
    /// An Asynchronous TCP Server that makes use of system managed threads
    /// and callbacks to stop the server ever locking up.
    /// </summary>
    public class NetworkEventreceiver
    {
        private readonly List<Client> clients;
        private readonly TcpListener tcpListener;
        private const string MAGIC_KEY = "quintessence\n\r";
        private const string COOKIE = "ABCD";
        private const string PASSWORD = "password";
        private string buffer;
        private string addr;

        public delegate void NetworkEventHandler(object sender, NetworkEventArgs e);
        public event NetworkEventHandler OnNetworkEvent;

        /// <summary> Gets or sets the device. </summary>
        /// <value> The device. </value>
        public Device Device { get; set; }

        /// <summary> Returns a <see cref="System.String" /> that represents this instance. </summary>
        /// <returns> A <see cref="System.String" /> that represents this instance. </returns>
        public override string ToString()
        {
            return addr;
        }

        public class NetworkEventArgs : EventArgs
        {
            public string device;
            public string ntwEvent;
            public string ntwPayload;

            public NetworkEventArgs(string device, string ntwEvent, string ntwPayload)
            {
                this.device = device;
                this.ntwEvent = ntwEvent;
                this.ntwPayload = ntwPayload;
            }
        }

        /// <summary>
        /// Internal class to join the TCP client and buffer together 
        /// for easy management in the server
        /// </summary>
        private class Client
        {
            /// <summary>
            /// Constructor for a new Client
            /// </summary>
            /// <param name="tcpClient">The TCP client</param>
            /// <param name="buffer">The byte array buffer</param>
            public Client(TcpClient tcpClient, byte[] buffer)
            {
                if (tcpClient == null) throw new ArgumentNullException("tcpClient");
                if (buffer == null) throw new ArgumentNullException("buffer");
                TcpClient = tcpClient;
                Buffer = buffer;
            }

            /// <summary>
            /// Gets the TCP Client
            /// </summary>
            public TcpClient TcpClient { get; private set; }

            /// <summary>
            /// Gets the Buffer.
            /// </summary>
            public byte[] Buffer { get; private set; }

            /// <summary>
            ///  Gets or sets the state.
            /// </summary>
            /// <value>
            /// The state.
            /// </value>
            public int State { get; set; }

            /// <summary>
            /// Gets the network stream
            /// </summary>
            public NetworkStream NetworkStream
            {
                get { return TcpClient.GetStream(); }
            }
        }
        /// <summary>
        /// Constructor for a new server using an IPAddress and Port
        /// </summary>
        /// <param name="localaddr">The Local IP Address for the server.</param>
        /// <param name="port">The port for the server.</param>
        public NetworkEventreceiver(IPAddress localaddr, int port)
        {
            Encoding = Encoding.Default;
            clients = new List<Client>();
            tcpListener = new TcpListener(localaddr, port);
            addr = localaddr + ":" + port;
        }

        /// <summary>
        /// The encoding to use when sending / receiving strings.
        /// </summary>
        public Encoding Encoding { get; set; }

        /// <summary>
        /// An enumerable collection of all the currently connected tcp clients
        /// </summary>
        public IEnumerable<TcpClient> TcpClients
        {
            get
            {
                foreach (Client client in clients)
                {
                    yield return client.TcpClient;
                }
            }
        }

        /// <summary>
        /// Starts the TCP Server listening for new clients.
        /// </summary>
        public void Start()
        {
            tcpListener.Start();
            tcpListener.BeginAcceptTcpClient(AcceptTcpClientCallback, null);
        }

        /// <summary>
        /// Stops the TCP Server listening for new clients and disconnects
        /// any currently connected clients.
        /// </summary>
        public void Stop()
        {
            lock (clients)
            {
                foreach (Client client in clients)
                {
                    client.TcpClient.Client.Disconnect(false);
                }
                clients.Clear();
            }
            tcpListener.Stop();
        }

        /// <summary>
        /// Writes a string to a given TCP Client
        /// </summary>
        /// <param name="tcpClient">The client to write to</param>
        /// <param name="data">The string to send.</param>
        public void Write(TcpClient tcpClient, string data)
        {
            byte[] bytes = Encoding.GetBytes(data);
            Write(tcpClient, bytes);
        }

        /// <summary>
        /// Writes a string to all clients connected.
        /// </summary>
        /// <param name="data">The string to send.</param>
        public void Write(string data)
        {
            foreach (Client client in clients)
            {
                Write(client.TcpClient, data);
            }
        }

        /// <summary>
        /// Writes a byte array to all clients connected.
        /// </summary>
        /// <param name="bytes">The bytes to send.</param>
        public void Write(byte[] bytes)
        {
            foreach (Client client in clients)
            {
                Write(client.TcpClient, bytes);
            }
        }

        /// <summary>
        /// Writes a byte array to a given TCP Client
        /// </summary>
        /// <param name="tcpClient">The client to write to</param>
        /// <param name="bytes">The bytes to send</param>
        public void Write(TcpClient tcpClient, byte[] bytes)
        {
            NetworkStream networkStream = tcpClient.GetStream();
            networkStream.BeginWrite(bytes, 0, bytes.Length, WriteCallback, tcpClient);
        }

        /// <summary>
        /// Callback for the write opertaion.
        /// </summary>
        /// <param name="result">The async result object</param>
        private void WriteCallback(IAsyncResult result)
        {
            var tcpClient = result.AsyncState as TcpClient;
            if (tcpClient != null)
            {
                NetworkStream networkStream = tcpClient.GetStream();
                networkStream.EndWrite(result);
            }
        }

        /// <summary>
        /// Callback for the accept tcp client opertaion.
        /// </summary>
        /// <param name="result">The async result object</param>
        private void AcceptTcpClientCallback(IAsyncResult result)
        {
            try
            {
                TcpClient tcpClient = tcpListener.EndAcceptTcpClient(result);
                var recvBuffer = new byte[tcpClient.ReceiveBufferSize];
                var client = new Client(tcpClient, recvBuffer) { State = 0 };
                lock (clients)
                {
                    clients.Add(client);
                }
                NetworkStream networkStream = client.NetworkStream;
                networkStream.BeginRead(client.Buffer, 0, client.Buffer.Length, ReadCallback, client);
                tcpListener.BeginAcceptTcpClient(AcceptTcpClientCallback, null);
            }
            catch (ObjectDisposedException)
            {
                Debug.WriteLine("Listen canceled.");
            }
        }

        /// <summary>
        /// Callback for the read opertaion.
        /// </summary>
        /// <param name="result">The async result object</param>
        private void ReadCallback(IAsyncResult result)
        {
            var client = result.AsyncState as Client;
            if (client == null) return;
            NetworkStream networkStream = client.NetworkStream;
            int read = networkStream.EndRead(result);
            if (read == 0)
            {
                lock (clients)
                {
                    ParseNetworkEvent(buffer);
                    clients.Remove(client);
                    return;
                }
            }

            string data = Encoding.GetString(client.Buffer, 0, read);
            
            switch (client.State)
            {
                case 0:
                    if (data == MAGIC_KEY)
                    {
                        Write(client.TcpClient, Encoding.ASCII.GetBytes(COOKIE + "\n"));
                        client.State = 1;
                    }
                    break;
                case 1:
                    const string token = COOKIE + ":" + PASSWORD;
                    string hash = HashString(token) + "\n";
                    if (data.ToUpper() == hash.ToUpper())
                    {
                        Write(client.TcpClient, Encoding.ASCII.GetBytes("accept\n"));
                        buffer = String.Empty;
                        client.State = 2;
                    }
                    break;
                case 2:

                    buffer += data;
       
                    break;
            }
            networkStream.BeginRead(client.Buffer, 0, client.Buffer.Length, ReadCallback, client);
        }

        private void ParseNetworkEvent (string networkEvent)
        {
            string ntwEvent;
            string ntwPayload = "";

            string[] split = networkEvent.Split('\n');

            if (split[0].StartsWith("payload withoutRelease"))
            {
                ntwEvent = split[1];
            }
            else
            {
                if (split[0].StartsWith("payload "))
                    ntwPayload = split[0].Remove(0, "payload ".Length);

                ntwEvent = split[2];
            }

            if (OnNetworkEvent != null)
                OnNetworkEvent(this, new NetworkEventArgs(Device.Manufacturer + " " + Device.Model, ntwEvent, ntwPayload));
        }

        /// <summary> Hashes the string. </summary>
        /// <param name="value">The value.</param>
        /// <returns></returns>
        private static string HashString(string value)
        {

            byte[] data = new MD5CryptoServiceProvider().ComputeHash(Encoding.ASCII.GetBytes(value));

            StringBuilder hashedString = new StringBuilder();

            for (int i = 0; i < data.Length; i++)

                hashedString.Append(data[i].ToString("X2"));

            return hashedString.ToString();

        }
    }
}