﻿using System;
using System.Collections.Generic;
using System.Diagnostics;
using System.IO;
using System.Net;
using System.Net.Sockets;
using System.Text;
using EGRemoteStudio.Protocol;
using Newtonsoft.Json;

namespace EGRemoteStudio
{
    /// <summary>
    /// An Asynchronous TCP Server that makes use of system managed threads
    /// and callbacks to stop the server ever locking up.
    /// </summary>
    public class RegAsyncTcpServer
    {
        private readonly List<Client> clients;
        private readonly TcpListener tcpListener;

        public delegate void RegisterEventHandler(object sender, DeviceRegisterArgs e);
        public event RegisterEventHandler OnDeviceRegister;

        public class DeviceRegisterArgs : EventArgs
        {
            public string regId;
            public string model;
            public string manufacturer;
            public int appVersion;

            public DeviceRegisterArgs(string regId, string model, string manufacturer, int appVersion)
            {
                this.regId = regId;
                this.model = model;
                this.manufacturer = manufacturer;
                this.appVersion = appVersion;
            }
        }

        /// <summary>
        /// Internal class to join the TCP client and buffer together 
        /// for easy management in the server
        /// </summary>
        private class Client
        {
            /// <summary>
            /// Constructor for a new Client
            /// </summary>
            /// <param name="tcpClient">The TCP client</param>
            /// <param name="buffer">The byte array buffer</param>
            public Client(TcpClient tcpClient, byte[] buffer)
            {
                if (tcpClient == null) throw new ArgumentNullException("tcpClient");
                if (buffer == null) throw new ArgumentNullException("buffer");
                TcpClient = tcpClient;
                Buffer = buffer;
            }

            /// <summary>
            /// Gets the TCP Client
            /// </summary>
            public TcpClient TcpClient { get; private set; }

            /// <summary>
            /// Gets the Buffer.
            /// </summary>
            public byte[] Buffer { get; private set; }

            /// <summary>
            /// Gets the network stream
            /// </summary>
            public NetworkStream NetworkStream
            {
                get { return TcpClient.GetStream(); }
            }
        }
        /// <summary>
        /// Constructor for a new server using an IPAddress and Port
        /// </summary>
        /// <param name="localaddr">The Local IP Address for the server.</param>
        /// <param name="port">The port for the server.</param>
        public RegAsyncTcpServer(IPAddress localaddr, int port)
            : this()
        {
            tcpListener = new TcpListener(localaddr, port);
        }

        /// <summary>
        /// Constructor for a new server using an end point
        /// </summary>
        /// <param name="localEP">The local end point for the server.</param>
        public RegAsyncTcpServer(IPEndPoint localEP)
            : this()
        {
            tcpListener = new TcpListener(localEP);
        }

        /// <summary>
        /// Private constructor for the common constructor operations.
        /// </summary>
        private RegAsyncTcpServer()
        {
            Encoding = Encoding.Default;
            clients = new List<Client>();
        }

        /// <summary>
        /// The encoding to use when sending / receiving strings.
        /// </summary>
        public Encoding Encoding { get; set; }

        /// <summary>
        /// An enumerable collection of all the currently connected tcp clients
        /// </summary>
        public IEnumerable<TcpClient> TcpClients
        {
            get
            {
                foreach (Client client in clients)
                {
                    yield return client.TcpClient;
                }
            }
        }

        /// <summary>
        /// Starts the TCP Server listening for new clients.
        /// </summary>
        public void Start()
        {
            tcpListener.Start();
            tcpListener.BeginAcceptTcpClient(AcceptTcpClientCallback, null);
        }

        /// <summary>
        /// Stops the TCP Server listening for new clients and disconnects
        /// any currently connected clients.
        /// </summary>
        public void Stop()
        {
            lock (clients)
            {
                foreach (Client client in clients)
                {
                    client.TcpClient.Client.Disconnect(false);
                }
                clients.Clear();
            }
            tcpListener.Stop();
        }

        /// <summary>
        /// Writes a string to a given TCP Client
        /// </summary>
        /// <param name="tcpClient">The client to write to</param>
        /// <param name="data">The string to send.</param>
        public void Write(TcpClient tcpClient, string data)
        {
            byte[] bytes = Encoding.GetBytes(data);
            Write(tcpClient, bytes);
        }

        /// <summary>
        /// Writes a string to all clients connected.
        /// </summary>
        /// <param name="data">The string to send.</param>
        public void Write(string data)
        {
            foreach (Client client in clients)
            {
                Write(client.TcpClient, data);
            }
        }

        /// <summary>
        /// Writes a byte array to all clients connected.
        /// </summary>
        /// <param name="bytes">The bytes to send.</param>
        public void Write(byte[] bytes)
        {
            foreach (Client client in clients)
            {
                Write(client.TcpClient, bytes);
            }
        }

        /// <summary>
        /// Writes a byte array to a given TCP Client
        /// </summary>
        /// <param name="tcpClient">The client to write to</param>
        /// <param name="bytes">The bytes to send</param>
        public void Write(TcpClient tcpClient, byte[] bytes)
        {
            NetworkStream networkStream = tcpClient.GetStream();
            networkStream.BeginWrite(bytes, 0, bytes.Length, WriteCallback, tcpClient);
        }

        /// <summary>
        /// Callback for the write opertaion.
        /// </summary>
        /// <param name="result">The async result object</param>
        private void WriteCallback(IAsyncResult result)
        {
            var tcpClient = result.AsyncState as TcpClient;
            if (tcpClient != null)
            {
                NetworkStream networkStream = tcpClient.GetStream();
                networkStream.EndWrite(result);
            }
        }

        /// <summary>
        /// Callback for the accept tcp client opertaion.
        /// </summary>
        /// <param name="result">The async result object</param>
        private void AcceptTcpClientCallback(IAsyncResult result)
        {
            try
            {
                TcpClient tcpClient = tcpListener.EndAcceptTcpClient(result);
                var buffer = new byte[tcpClient.ReceiveBufferSize];
                var client = new Client(tcpClient, buffer);
                lock (clients)
                {
                    clients.Add(client);
                }
                NetworkStream networkStream = client.NetworkStream;
                networkStream.BeginRead(client.Buffer, 0, client.Buffer.Length, ReadCallback, client);
                tcpListener.BeginAcceptTcpClient(AcceptTcpClientCallback, null);
            }
            catch (ObjectDisposedException)
            {
                Debug.WriteLine("Listen canceled.");
            }
        }

        /// <summary>
        /// Callback for the read opertaion.
        /// </summary>
        /// <param name="result">The async result object</param>
        private void ReadCallback(IAsyncResult result)
        {
            var client = result.AsyncState as Client;
            if (client == null) return;
            NetworkStream networkStream = client.NetworkStream;
            int read = networkStream.EndRead(result);
            if (read == 0)
            {
                lock (clients)
                {
                    clients.Remove(client);
                    return;
                }
            }

            string data = this.Encoding.GetString(client.Buffer, 0, read);
            RegIdMessage regIdMessage = JsonConvert.DeserializeObject<RegIdMessage>(data);
            
            if (regIdMessage != null)
            {
                if (OnDeviceRegister != null)
                    OnDeviceRegister(this, new DeviceRegisterArgs(regIdMessage.regId,
                        regIdMessage.model ?? String.Empty,
                        regIdMessage.manufacturer ?? String.Empty,
                        regIdMessage.appVersion));
            }
            
            networkStream.BeginRead(client.Buffer, 0, client.Buffer.Length, ReadCallback, client);
        }
    }
}